2021-02-04 04:00:29

by Muchun Song

[permalink] [raw]
Subject: [PATCH v14 0/8] Free some vmemmap pages of HugeTLB page

Hi all,

This patch series will free some vmemmap pages(struct page structures)
associated with each hugetlbpage when preallocated to save memory.

In order to reduce the difficulty of the first version of code review.
From this version, we disable PMD/huge page mapping of vmemmap if this
feature was enabled. This accutualy eliminate a bunch of the complex code
doing page table manipulation. When this patch series is solid, we cam add
the code of vmemmap page table manipulation in the future.

The struct page structures (page structs) are used to describe a physical
page frame. By default, there is a one-to-one mapping from a page frame to
it's corresponding page struct.

The HugeTLB pages consist of multiple base page size pages and is supported
by many architectures. See hugetlbpage.rst in the Documentation directory
for more details. On the x86 architecture, HugeTLB pages of size 2MB and 1GB
are currently supported. Since the base page size on x86 is 4KB, a 2MB
HugeTLB page consists of 512 base pages and a 1GB HugeTLB page consists of
4096 base pages. For each base page, there is a corresponding page struct.

Within the HugeTLB subsystem, only the first 4 page structs are used to
contain unique information about a HugeTLB page. HUGETLB_CGROUP_MIN_ORDER
provides this upper limit. The only 'useful' information in the remaining
page structs is the compound_head field, and this field is the same for all
tail pages.

By removing redundant page structs for HugeTLB pages, memory can returned to
the buddy allocator for other uses.

When the system boot up, every 2M HugeTLB has 512 struct page structs which
size is 8 pages(sizeof(struct page) * 512 / PAGE_SIZE).

HugeTLB struct pages(8 pages) page frame(8 pages)
+-----------+ ---virt_to_page---> +-----------+ mapping to +-----------+
| | | 0 | -------------> | 0 |
| | +-----------+ +-----------+
| | | 1 | -------------> | 1 |
| | +-----------+ +-----------+
| | | 2 | -------------> | 2 |
| | +-----------+ +-----------+
| | | 3 | -------------> | 3 |
| | +-----------+ +-----------+
| | | 4 | -------------> | 4 |
| 2MB | +-----------+ +-----------+
| | | 5 | -------------> | 5 |
| | +-----------+ +-----------+
| | | 6 | -------------> | 6 |
| | +-----------+ +-----------+
| | | 7 | -------------> | 7 |
| | +-----------+ +-----------+
| |
| |
| |
+-----------+

The value of page->compound_head is the same for all tail pages. The first
page of page structs (page 0) associated with the HugeTLB page contains the 4
page structs necessary to describe the HugeTLB. The only use of the remaining
pages of page structs (page 1 to page 7) is to point to page->compound_head.
Therefore, we can remap pages 2 to 7 to page 1. Only 2 pages of page structs
will be used for each HugeTLB page. This will allow us to free the remaining
6 pages to the buddy allocator.

Here is how things look after remapping.

HugeTLB struct pages(8 pages) page frame(8 pages)
+-----------+ ---virt_to_page---> +-----------+ mapping to +-----------+
| | | 0 | -------------> | 0 |
| | +-----------+ +-----------+
| | | 1 | -------------> | 1 |
| | +-----------+ +-----------+
| | | 2 | ----------------^ ^ ^ ^ ^ ^
| | +-----------+ | | | | |
| | | 3 | ------------------+ | | | |
| | +-----------+ | | | |
| | | 4 | --------------------+ | | |
| 2MB | +-----------+ | | |
| | | 5 | ----------------------+ | |
| | +-----------+ | |
| | | 6 | ------------------------+ |
| | +-----------+ |
| | | 7 | --------------------------+
| | +-----------+
| |
| |
| |
+-----------+

When a HugeTLB is freed to the buddy system, we should allocate 6 pages for
vmemmap pages and restore the previous mapping relationship.

Apart from 2MB HugeTLB page, we also have 1GB HugeTLB page. It is similar
to the 2MB HugeTLB page. We also can use this approach to free the vmemmap
pages.

In this case, for the 1GB HugeTLB page, we can save 4094 pages. This is a
very substantial gain. On our server, run some SPDK/QEMU applications which
will use 1024GB hugetlbpage. With this feature enabled, we can save ~16GB
(1G hugepage)/~12GB (2MB hugepage) memory.

Because there are vmemmap page tables reconstruction on the freeing/allocating
path, it increases some overhead. Here are some overhead analysis.

1) Allocating 10240 2MB hugetlb pages.

a) With this patch series applied:
# time echo 10240 > /proc/sys/vm/nr_hugepages

real 0m0.166s
user 0m0.000s
sys 0m0.166s

# bpftrace -e 'kprobe:alloc_fresh_huge_page { @start[tid] = nsecs; } kretprobe:alloc_fresh_huge_page /@start[tid]/ { @latency = hist(nsecs - @start[tid]); delete(@start[tid]); }'
Attaching 2 probes...

@latency:
[8K, 16K) 8360 |@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@|
[16K, 32K) 1868 |@@@@@@@@@@@ |
[32K, 64K) 10 | |
[64K, 128K) 2 | |

b) Without this patch series:
# time echo 10240 > /proc/sys/vm/nr_hugepages

real 0m0.066s
user 0m0.000s
sys 0m0.066s

# bpftrace -e 'kprobe:alloc_fresh_huge_page { @start[tid] = nsecs; } kretprobe:alloc_fresh_huge_page /@start[tid]/ { @latency = hist(nsecs - @start[tid]); delete(@start[tid]); }'
Attaching 2 probes...

@latency:
[4K, 8K) 10176 |@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@|
[8K, 16K) 62 | |
[16K, 32K) 2 | |

Summarize: this feature is about ~2x slower than before.

2) Freeing 10240 2MB hugetlb pages.

a) With this patch series applied:
# time echo 0 > /proc/sys/vm/nr_hugepages

real 0m0.004s
user 0m0.000s
sys 0m0.002s

# bpftrace -e 'kprobe:__free_hugepage { @start[tid] = nsecs; } kretprobe:__free_hugepage /@start[tid]/ { @latency = hist(nsecs - @start[tid]); delete(@start[tid]); }'
Attaching 2 probes...

@latency:
[16K, 32K) 10240 |@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@|

b) Without this patch series:
# time echo 0 > /proc/sys/vm/nr_hugepages

real 0m0.077s
user 0m0.001s
sys 0m0.075s

# bpftrace -e 'kprobe:__free_hugepage { @start[tid] = nsecs; } kretprobe:__free_hugepage /@start[tid]/ { @latency = hist(nsecs - @start[tid]); delete(@start[tid]); }'
Attaching 2 probes...

@latency:
[4K, 8K) 9950 |@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@|
[8K, 16K) 287 |@ |
[16K, 32K) 3 | |

Summarize: The overhead of __free_hugepage is about ~2-4x slower than before.
But according to the allocation test above, I think that here is
also ~2x slower than before.

But why the 'real' time of patched is smaller than before? Because
In this patch series, the freeing hugetlb is asynchronous(through
kwoker).

Although the overhead has increased, the overhead is not significant. Like Mike
said, "However, remember that the majority of use cases create hugetlb pages at
or shortly after boot time and add them to the pool. So, additional overhead is
at pool creation time. There is no change to 'normal run time' operations of
getting a page from or returning a page to the pool (think page fault/unmap)".

Todo:
- Free all of the tail vmemmap pages
Now for the 2MB HugrTLB page, we only free 6 vmemmap pages. we really can
free 7 vmemmap pages. In this case, we can see 8 of the 512 struct page
structures has beed set PG_head flag. If we can adjust compound_head()
slightly and make compound_head() return the real head struct page when
the parameter is the tail struct page but with PG_head flag set.

In order to make the code evolution route clearer. This feature can can be
a separate patch after this patchset is solid.

- Support for other architectures (e.g. aarch64).
- Enable PMD/huge page mapping of vmemmap even if this feature was enabled.

Changelog in v13 -> v14:
- Refuse to free the HugeTLB page when the system is under memory pressure.
- Use GFP_ATOMIC to allocate vmemmap pages instead of GFP_KERNEL.
- Rebase to linux-next 20210202.
- Fix and add some comments for vmemmap_remap_free().

Thanks to Oscar, Mike, David H and David R's suggestions and review.

Changelog in v12 -> v13:
- Remove VM_WARN_ON_PAGE macro.
- Add more comments in vmemmap_pte_range() and vmemmap_remap_free().

Thanks to Oscar and Mike's suggestions and review.

Changelog in v11 -> v12:
- Move VM_WARN_ON_PAGE to a separate patch.
- Call __free_hugepage() with hugetlb_lock (See patch #5.) to serialize
with dissolve_free_huge_page(). It is to prepare for patch #9.
- Introduce PageHugeInflight. See patch #9.

Changelog in v10 -> v11:
- Fix compiler error when !CONFIG_HUGETLB_PAGE_FREE_VMEMMAP.
- Rework some comments and commit changes.
- Rework vmemmap_remap_free() to 3 parameters.

Thanks to Oscar and Mike's suggestions and review.

Changelog in v9 -> v10:
- Fix a bug in patch #11. Thanks to Oscar for pointing that out.
- Rework some commit log or comments. Thanks Mike and Oscar for the suggestions.
- Drop VMEMMAP_TAIL_PAGE_REUSE in the patch #3.

Thank you very much Mike and Oscar for reviewing the code.

Changelog in v8 -> v9:
- Rework some code. Very thanks to Oscar.
- Put all the non-hugetlb vmemmap functions under sparsemem-vmemmap.c.

Changelog in v7 -> v8:
- Adjust the order of patches.

Very thanks to David and Oscar. Your suggestions are very valuable.

Changelog in v6 -> v7:
- Rebase to linux-next 20201130
- Do not use basepage mapping for vmemmap when this feature is disabled.
- Rework some patchs.
[PATCH v6 08/16] mm/hugetlb: Free the vmemmap pages associated with each hugetlb page
[PATCH v6 10/16] mm/hugetlb: Allocate the vmemmap pages associated with each hugetlb page

Thanks to Oscar and Barry.

Changelog in v5 -> v6:
- Disable PMD/huge page mapping of vmemmap if this feature was enabled.
- Simplify the first version code.

Changelog in v4 -> v5:
- Rework somme comments and code in the [PATCH v4 04/21] and [PATCH v4 05/21].

Thanks to Mike and Oscar's suggestions.

Changelog in v3 -> v4:
- Move all the vmemmap functions to hugetlb_vmemmap.c.
- Make the CONFIG_HUGETLB_PAGE_FREE_VMEMMAP default to y, if we want to
disable this feature, we should disable it by a boot/kernel command line.
- Remove vmemmap_pgtable_{init, deposit, withdraw}() helper functions.
- Initialize page table lock for vmemmap through core_initcall mechanism.

Thanks for Mike and Oscar's suggestions.

Changelog in v2 -> v3:
- Rename some helps function name. Thanks Mike.
- Rework some code. Thanks Mike and Oscar.
- Remap the tail vmemmap page with PAGE_KERNEL_RO instead of PAGE_KERNEL.
Thanks Matthew.
- Add some overhead analysis in the cover letter.
- Use vmemap pmd table lock instead of a hugetlb specific global lock.

Changelog in v1 -> v2:
- Fix do not call dissolve_compound_page in alloc_huge_page_vmemmap().
- Fix some typo and code style problems.
- Remove unused handle_vmemmap_fault().
- Merge some commits to one commit suggested by Mike.

Muchun Song (8):
mm: memory_hotplug: factor out bootmem core functions to
bootmem_info.c
mm: hugetlb: introduce a new config HUGETLB_PAGE_FREE_VMEMMAP
mm: hugetlb: free the vmemmap pages associated with each HugeTLB page
mm: hugetlb: alloc the vmemmap pages associated with each HugeTLB page
mm: hugetlb: add a kernel parameter hugetlb_free_vmemmap
mm: hugetlb: introduce nr_free_vmemmap_pages in the struct hstate
mm: hugetlb: gather discrete indexes of tail page
mm: hugetlb: optimize the code with the help of the compiler

Documentation/admin-guide/kernel-parameters.txt | 14 ++
Documentation/admin-guide/mm/hugetlbpage.rst | 3 +
arch/x86/mm/init_64.c | 13 +-
fs/Kconfig | 6 +
include/linux/bootmem_info.h | 65 +++++
include/linux/hugetlb.h | 43 +++-
include/linux/hugetlb_cgroup.h | 19 +-
include/linux/memory_hotplug.h | 27 --
include/linux/mm.h | 5 +
mm/Makefile | 2 +
mm/bootmem_info.c | 124 ++++++++++
mm/hugetlb.c | 23 +-
mm/hugetlb_vmemmap.c | 314 ++++++++++++++++++++++++
mm/hugetlb_vmemmap.h | 33 +++
mm/memory_hotplug.c | 116 ---------
mm/sparse-vmemmap.c | 280 +++++++++++++++++++++
mm/sparse.c | 1 +
17 files changed, 930 insertions(+), 158 deletions(-)
create mode 100644 include/linux/bootmem_info.h
create mode 100644 mm/bootmem_info.c
create mode 100644 mm/hugetlb_vmemmap.c
create mode 100644 mm/hugetlb_vmemmap.h

--
2.11.0


2021-02-04 04:00:40

by Muchun Song

[permalink] [raw]
Subject: [PATCH v14 2/8] mm: hugetlb: introduce a new config HUGETLB_PAGE_FREE_VMEMMAP

The option HUGETLB_PAGE_FREE_VMEMMAP allows for the freeing of
some vmemmap pages associated with pre-allocated HugeTLB pages.
For example, on X86_64 6 vmemmap pages of size 4KB each can be
saved for each 2MB HugeTLB page. 4094 vmemmap pages of size 4KB
each can be saved for each 1GB HugeTLB page.

When a HugeTLB page is allocated or freed, the vmemmap array
representing the range associated with the page will need to be
remapped. When a page is allocated, vmemmap pages are freed
after remapping. When a page is freed, previously discarded
vmemmap pages must be allocated before remapping.

The config option is introduced early so that supporting code
can be written to depend on the option. The initial version of
the code only provides support for x86-64.

Like other code which frees vmemmap, this config option depends on
HAVE_BOOTMEM_INFO_NODE. The routine register_page_bootmem_info() is
used to register bootmem info. Therefore, make sure
register_page_bootmem_info is enabled if HUGETLB_PAGE_FREE_VMEMMAP
is defined.

Signed-off-by: Muchun Song <[email protected]>
Reviewed-by: Oscar Salvador <[email protected]>
Acked-by: Mike Kravetz <[email protected]>
---
arch/x86/mm/init_64.c | 2 +-
fs/Kconfig | 6 ++++++
2 files changed, 7 insertions(+), 1 deletion(-)

diff --git a/arch/x86/mm/init_64.c b/arch/x86/mm/init_64.c
index 0a45f062826e..0435bee2e172 100644
--- a/arch/x86/mm/init_64.c
+++ b/arch/x86/mm/init_64.c
@@ -1225,7 +1225,7 @@ static struct kcore_list kcore_vsyscall;

static void __init register_page_bootmem_info(void)
{
-#ifdef CONFIG_NUMA
+#if defined(CONFIG_NUMA) || defined(CONFIG_HUGETLB_PAGE_FREE_VMEMMAP)
int i;

for_each_online_node(i)
diff --git a/fs/Kconfig b/fs/Kconfig
index 97e7b77c9309..de87f234f1e9 100644
--- a/fs/Kconfig
+++ b/fs/Kconfig
@@ -237,6 +237,12 @@ config HUGETLBFS
config HUGETLB_PAGE
def_bool HUGETLBFS

+config HUGETLB_PAGE_FREE_VMEMMAP
+ def_bool HUGETLB_PAGE
+ depends on X86_64
+ depends on SPARSEMEM_VMEMMAP
+ depends on HAVE_BOOTMEM_INFO_NODE
+
config MEMFD_CREATE
def_bool TMPFS || HUGETLBFS

--
2.11.0

2021-02-04 04:01:08

by Muchun Song

[permalink] [raw]
Subject: [PATCH v14 7/8] mm: hugetlb: gather discrete indexes of tail page

For HugeTLB page, there are more metadata to save in the struct page.
But the head struct page cannot meet our needs, so we have to abuse
other tail struct page to store the metadata. In order to avoid
conflicts caused by subsequent use of more tail struct pages, we can
gather these discrete indexes of tail struct page. In this case, it
will be easier to add a new tail page index later.

There are only (RESERVE_VMEMMAP_SIZE / sizeof(struct page)) struct
page structs that can be used when CONFIG_HUGETLB_PAGE_FREE_VMEMMAP,
so add a BUILD_BUG_ON to catch invalid usage of the tail struct page.

Signed-off-by: Muchun Song <[email protected]>
Reviewed-by: Oscar Salvador <[email protected]>
---
include/linux/hugetlb.h | 20 ++++++++++++++++++--
include/linux/hugetlb_cgroup.h | 19 +++++++++++--------
mm/hugetlb_vmemmap.c | 8 ++++++++
3 files changed, 37 insertions(+), 10 deletions(-)

diff --git a/include/linux/hugetlb.h b/include/linux/hugetlb.h
index 775aea53669a..822ab2f5542a 100644
--- a/include/linux/hugetlb.h
+++ b/include/linux/hugetlb.h
@@ -28,6 +28,22 @@ typedef struct { unsigned long pd; } hugepd_t;
#include <linux/shm.h>
#include <asm/tlbflush.h>

+/*
+ * For HugeTLB page, there are more metadata to save in the struct page. But
+ * the head struct page cannot meet our needs, so we have to abuse other tail
+ * struct page to store the metadata. In order to avoid conflicts caused by
+ * subsequent use of more tail struct pages, we gather these discrete indexes
+ * of tail struct page here.
+ */
+enum {
+ SUBPAGE_INDEX_SUBPOOL = 1, /* reuse page->private */
+#ifdef CONFIG_CGROUP_HUGETLB
+ SUBPAGE_INDEX_CGROUP, /* reuse page->private */
+ SUBPAGE_INDEX_CGROUP_RSVD, /* reuse page->private */
+#endif
+ NR_USED_SUBPAGE,
+};
+
struct hugepage_subpool {
spinlock_t lock;
long count;
@@ -607,13 +623,13 @@ extern unsigned int default_hstate_idx;
*/
static inline struct hugepage_subpool *hugetlb_page_subpool(struct page *hpage)
{
- return (struct hugepage_subpool *)(hpage+1)->private;
+ return (void *)page_private(hpage + SUBPAGE_INDEX_SUBPOOL);
}

static inline void hugetlb_set_page_subpool(struct page *hpage,
struct hugepage_subpool *subpool)
{
- set_page_private(hpage+1, (unsigned long)subpool);
+ set_page_private(hpage + SUBPAGE_INDEX_SUBPOOL, (unsigned long)subpool);
}

static inline struct hstate *hstate_file(struct file *f)
diff --git a/mm/hugetlb_vmemmap.c b/mm/hugetlb_vmemmap.c
index 36ebd677e606..8efad9978821 100644
--- a/mm/hugetlb_vmemmap.c
+++ b/mm/hugetlb_vmemmap.c
@@ -272,6 +272,14 @@ void __init hugetlb_vmemmap_init(struct hstate *h)
unsigned int nr_pages = pages_per_huge_page(h);
unsigned int vmemmap_pages;

+ /*
+ * There are only (RESERVE_VMEMMAP_SIZE / sizeof(struct page)) struct
+ * page structs that can be used when CONFIG_HUGETLB_PAGE_FREE_VMEMMAP,
+ * so add a BUILD_BUG_ON to catch invalid usage of the tail struct page.
+ */
+ BUILD_BUG_ON(NR_USED_SUBPAGE >=
+ RESERVE_VMEMMAP_SIZE / sizeof(struct page));
+
if (!hugetlb_free_vmemmap_enabled)
return;

--
2.11.0

2021-02-04 04:02:54

by Muchun Song

[permalink] [raw]
Subject: [PATCH v14 3/8] mm: hugetlb: free the vmemmap pages associated with each HugeTLB page

Every HugeTLB has more than one struct page structure. We __know__ that
we only use the first 4(HUGETLB_CGROUP_MIN_ORDER) struct page structures
to store metadata associated with each HugeTLB.

There are a lot of struct page structures associated with each HugeTLB
page. For tail pages, the value of compound_head is the same. So we can
reuse first page of tail page structures. We map the virtual addresses
of the remaining pages of tail page structures to the first tail page
struct, and then free these page frames. Therefore, we need to reserve
two pages as vmemmap areas.

When we allocate a HugeTLB page from the buddy, we can free some vmemmap
pages associated with each HugeTLB page. It is more appropriate to do it
in the prep_new_huge_page().

The free_vmemmap_pages_per_hpage(), which indicates how many vmemmap
pages associated with a HugeTLB page can be freed, returns zero for
now, which means the feature is disabled. We will enable it once all
the infrastructure is there.

Signed-off-by: Muchun Song <[email protected]>
---
include/linux/bootmem_info.h | 27 +++++-
include/linux/mm.h | 3 +
mm/Makefile | 1 +
mm/hugetlb.c | 3 +
mm/hugetlb_vmemmap.c | 219 +++++++++++++++++++++++++++++++++++++++++++
mm/hugetlb_vmemmap.h | 20 ++++
mm/sparse-vmemmap.c | 207 ++++++++++++++++++++++++++++++++++++++++
7 files changed, 479 insertions(+), 1 deletion(-)
create mode 100644 mm/hugetlb_vmemmap.c
create mode 100644 mm/hugetlb_vmemmap.h

diff --git a/include/linux/bootmem_info.h b/include/linux/bootmem_info.h
index 4ed6dee1adc9..ec03a624dfa2 100644
--- a/include/linux/bootmem_info.h
+++ b/include/linux/bootmem_info.h
@@ -2,7 +2,7 @@
#ifndef __LINUX_BOOTMEM_INFO_H
#define __LINUX_BOOTMEM_INFO_H

-#include <linux/mmzone.h>
+#include <linux/mm.h>

/*
* Types for free bootmem stored in page->lru.next. These have to be in
@@ -22,6 +22,27 @@ void __init register_page_bootmem_info_node(struct pglist_data *pgdat);
void get_page_bootmem(unsigned long info, struct page *page,
unsigned long type);
void put_page_bootmem(struct page *page);
+
+/*
+ * Any memory allocated via the memblock allocator and not via the
+ * buddy will be marked reserved already in the memmap. For those
+ * pages, we can call this function to free it to buddy allocator.
+ */
+static inline void free_bootmem_page(struct page *page)
+{
+ unsigned long magic = (unsigned long)page->freelist;
+
+ /*
+ * The reserve_bootmem_region sets the reserved flag on bootmem
+ * pages.
+ */
+ VM_BUG_ON_PAGE(page_ref_count(page) != 2, page);
+
+ if (magic == SECTION_INFO || magic == MIX_SECTION_INFO)
+ put_page_bootmem(page);
+ else
+ VM_BUG_ON_PAGE(1, page);
+}
#else
static inline void register_page_bootmem_info_node(struct pglist_data *pgdat)
{
@@ -35,6 +56,10 @@ static inline void get_page_bootmem(unsigned long info, struct page *page,
unsigned long type)
{
}
+
+static inline void free_bootmem_page(struct page *page)
+{
+}
#endif

#endif /* __LINUX_BOOTMEM_INFO_H */
diff --git a/include/linux/mm.h b/include/linux/mm.h
index a608feb0d42e..d7dddf334779 100644
--- a/include/linux/mm.h
+++ b/include/linux/mm.h
@@ -2979,6 +2979,9 @@ static inline void print_vma_addr(char *prefix, unsigned long rip)
}
#endif

+void vmemmap_remap_free(unsigned long start, unsigned long end,
+ unsigned long reuse);
+
void *sparse_buffer_alloc(unsigned long size);
struct page * __populate_section_memmap(unsigned long pfn,
unsigned long nr_pages, int nid, struct vmem_altmap *altmap);
diff --git a/mm/Makefile b/mm/Makefile
index ce4ddbe4461d..47b250e4c9b2 100644
--- a/mm/Makefile
+++ b/mm/Makefile
@@ -71,6 +71,7 @@ obj-$(CONFIG_FRONTSWAP) += frontswap.o
obj-$(CONFIG_ZSWAP) += zswap.o
obj-$(CONFIG_HAS_DMA) += dmapool.o
obj-$(CONFIG_HUGETLBFS) += hugetlb.o
+obj-$(CONFIG_HUGETLB_PAGE_FREE_VMEMMAP) += hugetlb_vmemmap.o
obj-$(CONFIG_NUMA) += mempolicy.o
obj-$(CONFIG_SPARSEMEM) += sparse.o
obj-$(CONFIG_SPARSEMEM_VMEMMAP) += sparse-vmemmap.o
diff --git a/mm/hugetlb.c b/mm/hugetlb.c
index 8c53f3f2e12e..4cfca27c6d32 100644
--- a/mm/hugetlb.c
+++ b/mm/hugetlb.c
@@ -42,6 +42,7 @@
#include <linux/userfaultfd_k.h>
#include <linux/page_owner.h>
#include "internal.h"
+#include "hugetlb_vmemmap.h"

int hugetlb_max_hstate __read_mostly;
unsigned int default_hstate_idx;
@@ -1462,6 +1463,8 @@ void free_huge_page(struct page *page)

static void prep_new_huge_page(struct hstate *h, struct page *page, int nid)
{
+ free_huge_page_vmemmap(h, page);
+
INIT_LIST_HEAD(&page->lru);
set_compound_page_dtor(page, HUGETLB_PAGE_DTOR);
set_hugetlb_cgroup(page, NULL);
diff --git a/mm/hugetlb_vmemmap.c b/mm/hugetlb_vmemmap.c
new file mode 100644
index 000000000000..ddd872ab6180
--- /dev/null
+++ b/mm/hugetlb_vmemmap.c
@@ -0,0 +1,219 @@
+// SPDX-License-Identifier: GPL-2.0
+/*
+ * Free some vmemmap pages of HugeTLB
+ *
+ * Copyright (c) 2020, Bytedance. All rights reserved.
+ *
+ * Author: Muchun Song <[email protected]>
+ *
+ * The struct page structures (page structs) are used to describe a physical
+ * page frame. By default, there is a one-to-one mapping from a page frame to
+ * it's corresponding page struct.
+ *
+ * HugeTLB pages consist of multiple base page size pages and is supported by
+ * many architectures. See hugetlbpage.rst in the Documentation directory for
+ * more details. On the x86-64 architecture, HugeTLB pages of size 2MB and 1GB
+ * are currently supported. Since the base page size on x86 is 4KB, a 2MB
+ * HugeTLB page consists of 512 base pages and a 1GB HugeTLB page consists of
+ * 4096 base pages. For each base page, there is a corresponding page struct.
+ *
+ * Within the HugeTLB subsystem, only the first 4 page structs are used to
+ * contain unique information about a HugeTLB page. HUGETLB_CGROUP_MIN_ORDER
+ * provides this upper limit. The only 'useful' information in the remaining
+ * page structs is the compound_head field, and this field is the same for all
+ * tail pages.
+ *
+ * By removing redundant page structs for HugeTLB pages, memory can be returned
+ * to the buddy allocator for other uses.
+ *
+ * Different architectures support different HugeTLB pages. For example, the
+ * following table is the HugeTLB page size supported by x86 and arm64
+ * architectures. Because arm64 supports 4k, 16k, and 64k base pages and
+ * supports contiguous entries, so it supports many kinds of sizes of HugeTLB
+ * page.
+ *
+ * +--------------+-----------+-----------------------------------------------+
+ * | Architecture | Page Size | HugeTLB Page Size |
+ * +--------------+-----------+-----------+-----------+-----------+-----------+
+ * | x86-64 | 4KB | 2MB | 1GB | | |
+ * +--------------+-----------+-----------+-----------+-----------+-----------+
+ * | | 4KB | 64KB | 2MB | 32MB | 1GB |
+ * | +-----------+-----------+-----------+-----------+-----------+
+ * | arm64 | 16KB | 2MB | 32MB | 1GB | |
+ * | +-----------+-----------+-----------+-----------+-----------+
+ * | | 64KB | 2MB | 512MB | 16GB | |
+ * +--------------+-----------+-----------+-----------+-----------+-----------+
+ *
+ * When the system boot up, every HugeTLB page has more than one struct page
+ * structs which size is (unit: pages):
+ *
+ * struct_size = HugeTLB_Size / PAGE_SIZE * sizeof(struct page) / PAGE_SIZE
+ *
+ * Where HugeTLB_Size is the size of the HugeTLB page. We know that the size
+ * of the HugeTLB page is always n times PAGE_SIZE. So we can get the following
+ * relationship.
+ *
+ * HugeTLB_Size = n * PAGE_SIZE
+ *
+ * Then,
+ *
+ * struct_size = n * PAGE_SIZE / PAGE_SIZE * sizeof(struct page) / PAGE_SIZE
+ * = n * sizeof(struct page) / PAGE_SIZE
+ *
+ * We can use huge mapping at the pud/pmd level for the HugeTLB page.
+ *
+ * For the HugeTLB page of the pmd level mapping, then
+ *
+ * struct_size = n * sizeof(struct page) / PAGE_SIZE
+ * = PAGE_SIZE / sizeof(pte_t) * sizeof(struct page) / PAGE_SIZE
+ * = sizeof(struct page) / sizeof(pte_t)
+ * = 64 / 8
+ * = 8 (pages)
+ *
+ * Where n is how many pte entries which one page can contains. So the value of
+ * n is (PAGE_SIZE / sizeof(pte_t)).
+ *
+ * This optimization only supports 64-bit system, so the value of sizeof(pte_t)
+ * is 8. And this optimization also applicable only when the size of struct page
+ * is a power of two. In most cases, the size of struct page is 64 bytes (e.g.
+ * x86-64 and arm64). So if we use pmd level mapping for a HugeTLB page, the
+ * size of struct page structs of it is 8 page frames which size depends on the
+ * size of the base page.
+ *
+ * For the HugeTLB page of the pud level mapping, then
+ *
+ * struct_size = PAGE_SIZE / sizeof(pmd_t) * struct_size(pmd)
+ * = PAGE_SIZE / 8 * 8 (pages)
+ * = PAGE_SIZE (pages)
+ *
+ * Where the struct_size(pmd) is the size of the struct page structs of a
+ * HugeTLB page of the pmd level mapping.
+ *
+ * E.g.: A 2MB HugeTLB page on x86_64 consists in 8 page frames while 1GB
+ * HugeTLB page consists in 4096.
+ *
+ * Next, we take the pmd level mapping of the HugeTLB page as an example to
+ * show the internal implementation of this optimization. There are 8 pages
+ * struct page structs associated with a HugeTLB page which is pmd mapped.
+ *
+ * Here is how things look before optimization.
+ *
+ * HugeTLB struct pages(8 pages) page frame(8 pages)
+ * +-----------+ ---virt_to_page---> +-----------+ mapping to +-----------+
+ * | | | 0 | -------------> | 0 |
+ * | | +-----------+ +-----------+
+ * | | | 1 | -------------> | 1 |
+ * | | +-----------+ +-----------+
+ * | | | 2 | -------------> | 2 |
+ * | | +-----------+ +-----------+
+ * | | | 3 | -------------> | 3 |
+ * | | +-----------+ +-----------+
+ * | | | 4 | -------------> | 4 |
+ * | PMD | +-----------+ +-----------+
+ * | level | | 5 | -------------> | 5 |
+ * | mapping | +-----------+ +-----------+
+ * | | | 6 | -------------> | 6 |
+ * | | +-----------+ +-----------+
+ * | | | 7 | -------------> | 7 |
+ * | | +-----------+ +-----------+
+ * | |
+ * | |
+ * | |
+ * +-----------+
+ *
+ * The value of page->compound_head is the same for all tail pages. The first
+ * page of page structs (page 0) associated with the HugeTLB page contains the 4
+ * page structs necessary to describe the HugeTLB. The only use of the remaining
+ * pages of page structs (page 1 to page 7) is to point to page->compound_head.
+ * Therefore, we can remap pages 2 to 7 to page 1. Only 2 pages of page structs
+ * will be used for each HugeTLB page. This will allow us to free the remaining
+ * 6 pages to the buddy allocator.
+ *
+ * Here is how things look after remapping.
+ *
+ * HugeTLB struct pages(8 pages) page frame(8 pages)
+ * +-----------+ ---virt_to_page---> +-----------+ mapping to +-----------+
+ * | | | 0 | -------------> | 0 |
+ * | | +-----------+ +-----------+
+ * | | | 1 | -------------> | 1 |
+ * | | +-----------+ +-----------+
+ * | | | 2 | ----------------^ ^ ^ ^ ^ ^
+ * | | +-----------+ | | | | |
+ * | | | 3 | ------------------+ | | | |
+ * | | +-----------+ | | | |
+ * | | | 4 | --------------------+ | | |
+ * | PMD | +-----------+ | | |
+ * | level | | 5 | ----------------------+ | |
+ * | mapping | +-----------+ | |
+ * | | | 6 | ------------------------+ |
+ * | | +-----------+ |
+ * | | | 7 | --------------------------+
+ * | | +-----------+
+ * | |
+ * | |
+ * | |
+ * +-----------+
+ *
+ * When a HugeTLB is freed to the buddy system, we should allocate 6 pages for
+ * vmemmap pages and restore the previous mapping relationship.
+ *
+ * For the HugeTLB page of the pud level mapping. It is similar to the former.
+ * We also can use this approach to free (PAGE_SIZE - 2) vmemmap pages.
+ *
+ * Apart from the HugeTLB page of the pmd/pud level mapping, some architectures
+ * (e.g. aarch64) provides a contiguous bit in the translation table entries
+ * that hints to the MMU to indicate that it is one of a contiguous set of
+ * entries that can be cached in a single TLB entry.
+ *
+ * The contiguous bit is used to increase the mapping size at the pmd and pte
+ * (last) level. So this type of HugeTLB page can be optimized only when its
+ * size of the struct page structs is greater than 2 pages.
+ */
+#include "hugetlb_vmemmap.h"
+
+/*
+ * There are a lot of struct page structures associated with each HugeTLB page.
+ * For tail pages, the value of compound_head is the same. So we can reuse first
+ * page of tail page structures. We map the virtual addresses of the remaining
+ * pages of tail page structures to the first tail page struct, and then free
+ * these page frames. Therefore, we need to reserve two pages as vmemmap areas.
+ */
+#define RESERVE_VMEMMAP_NR 2U
+#define RESERVE_VMEMMAP_SIZE (RESERVE_VMEMMAP_NR << PAGE_SHIFT)
+
+/*
+ * How many vmemmap pages associated with a HugeTLB page that can be freed
+ * to the buddy allocator.
+ *
+ * Todo: Returns zero for now, which means the feature is disabled. We will
+ * enable it once all the infrastructure is there.
+ */
+static inline unsigned int free_vmemmap_pages_per_hpage(struct hstate *h)
+{
+ return 0;
+}
+
+static inline unsigned long free_vmemmap_pages_size_per_hpage(struct hstate *h)
+{
+ return (unsigned long)free_vmemmap_pages_per_hpage(h) << PAGE_SHIFT;
+}
+
+void free_huge_page_vmemmap(struct hstate *h, struct page *head)
+{
+ unsigned long vmemmap_addr = (unsigned long)head;
+ unsigned long vmemmap_end, vmemmap_reuse;
+
+ if (!free_vmemmap_pages_per_hpage(h))
+ return;
+
+ vmemmap_addr += RESERVE_VMEMMAP_SIZE;
+ vmemmap_end = vmemmap_addr + free_vmemmap_pages_size_per_hpage(h);
+ vmemmap_reuse = vmemmap_addr - PAGE_SIZE;
+
+ /*
+ * Remap the vmemmap virtual address range [@vmemmap_addr, @vmemmap_end)
+ * to the page which @vmemmap_reuse is mapped to, then free the vmemmap
+ * pages which the range are mapped to.
+ */
+ vmemmap_remap_free(vmemmap_addr, vmemmap_end, vmemmap_reuse);
+}
diff --git a/mm/hugetlb_vmemmap.h b/mm/hugetlb_vmemmap.h
new file mode 100644
index 000000000000..6923f03534d5
--- /dev/null
+++ b/mm/hugetlb_vmemmap.h
@@ -0,0 +1,20 @@
+// SPDX-License-Identifier: GPL-2.0
+/*
+ * Free some vmemmap pages of HugeTLB
+ *
+ * Copyright (c) 2020, Bytedance. All rights reserved.
+ *
+ * Author: Muchun Song <[email protected]>
+ */
+#ifndef _LINUX_HUGETLB_VMEMMAP_H
+#define _LINUX_HUGETLB_VMEMMAP_H
+#include <linux/hugetlb.h>
+
+#ifdef CONFIG_HUGETLB_PAGE_FREE_VMEMMAP
+void free_huge_page_vmemmap(struct hstate *h, struct page *head);
+#else
+static inline void free_huge_page_vmemmap(struct hstate *h, struct page *head)
+{
+}
+#endif /* CONFIG_HUGETLB_PAGE_FREE_VMEMMAP */
+#endif /* _LINUX_HUGETLB_VMEMMAP_H */
diff --git a/mm/sparse-vmemmap.c b/mm/sparse-vmemmap.c
index 16183d85a7d5..50c1dc00b686 100644
--- a/mm/sparse-vmemmap.c
+++ b/mm/sparse-vmemmap.c
@@ -27,8 +27,215 @@
#include <linux/spinlock.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
+#include <linux/pgtable.h>
+#include <linux/bootmem_info.h>
+
#include <asm/dma.h>
#include <asm/pgalloc.h>
+#include <asm/tlbflush.h>
+
+/**
+ * vmemmap_remap_walk - walk vmemmap page table
+ *
+ * @remap_pte: called for each non-empty PTE (lowest-level) entry.
+ * @reuse_page: the page which is reused for the tail vmemmap pages.
+ * @reuse_addr: the virtual address of the @reuse_page page.
+ * @vmemmap_pages: the list head of the vmemmap pages that can be freed.
+ */
+struct vmemmap_remap_walk {
+ void (*remap_pte)(pte_t *pte, unsigned long addr,
+ struct vmemmap_remap_walk *walk);
+ struct page *reuse_page;
+ unsigned long reuse_addr;
+ struct list_head *vmemmap_pages;
+};
+
+static void vmemmap_pte_range(pmd_t *pmd, unsigned long addr,
+ unsigned long end,
+ struct vmemmap_remap_walk *walk)
+{
+ pte_t *pte;
+
+ pte = pte_offset_kernel(pmd, addr);
+
+ /*
+ * The reuse_page is found 'first' in table walk before we start
+ * remapping (which is calling @walk->remap_pte).
+ */
+ if (!walk->reuse_page) {
+ BUG_ON(pte_none(*pte) || walk->reuse_addr != addr);
+
+ walk->reuse_page = pte_page(*pte++);
+ /*
+ * Because the reuse address is part of the range that we are
+ * walking, skip the reuse address range.
+ */
+ addr += PAGE_SIZE;
+ }
+
+ for (; addr != end; addr += PAGE_SIZE, pte++) {
+ BUG_ON(pte_none(*pte));
+
+ walk->remap_pte(pte, addr, walk);
+ }
+}
+
+static void vmemmap_pmd_range(pud_t *pud, unsigned long addr,
+ unsigned long end,
+ struct vmemmap_remap_walk *walk)
+{
+ pmd_t *pmd;
+ unsigned long next;
+
+ pmd = pmd_offset(pud, addr);
+ do {
+ BUG_ON(pmd_none(*pmd) || pmd_leaf(*pmd));
+
+ next = pmd_addr_end(addr, end);
+ vmemmap_pte_range(pmd, addr, next, walk);
+ } while (pmd++, addr = next, addr != end);
+}
+
+static void vmemmap_pud_range(p4d_t *p4d, unsigned long addr,
+ unsigned long end,
+ struct vmemmap_remap_walk *walk)
+{
+ pud_t *pud;
+ unsigned long next;
+
+ pud = pud_offset(p4d, addr);
+ do {
+ BUG_ON(pud_none(*pud));
+
+ next = pud_addr_end(addr, end);
+ vmemmap_pmd_range(pud, addr, next, walk);
+ } while (pud++, addr = next, addr != end);
+}
+
+static void vmemmap_p4d_range(pgd_t *pgd, unsigned long addr,
+ unsigned long end,
+ struct vmemmap_remap_walk *walk)
+{
+ p4d_t *p4d;
+ unsigned long next;
+
+ p4d = p4d_offset(pgd, addr);
+ do {
+ BUG_ON(p4d_none(*p4d));
+
+ next = p4d_addr_end(addr, end);
+ vmemmap_pud_range(p4d, addr, next, walk);
+ } while (p4d++, addr = next, addr != end);
+}
+
+static void vmemmap_remap_range(unsigned long start, unsigned long end,
+ struct vmemmap_remap_walk *walk)
+{
+ unsigned long addr = start;
+ unsigned long next;
+ pgd_t *pgd;
+
+ VM_BUG_ON(!IS_ALIGNED(start, PAGE_SIZE));
+ VM_BUG_ON(!IS_ALIGNED(end, PAGE_SIZE));
+
+ pgd = pgd_offset_k(addr);
+ do {
+ BUG_ON(pgd_none(*pgd));
+
+ next = pgd_addr_end(addr, end);
+ vmemmap_p4d_range(pgd, addr, next, walk);
+ } while (pgd++, addr = next, addr != end);
+
+ /*
+ * We do not change the mapping of the vmemmap virtual address range
+ * [@start, @start + PAGE_SIZE) which belongs to the reuse range.
+ * So we not need to flush the TLB.
+ */
+ flush_tlb_kernel_range(start + PAGE_SIZE, end);
+}
+
+/*
+ * Free a vmemmap page. A vmemmap page can be allocated from the memblock
+ * allocator or buddy allocator. If the PG_reserved flag is set, it means
+ * that it allocated from the memblock allocator, just free it via the
+ * free_bootmem_page(). Otherwise, use __free_page().
+ */
+static inline void free_vmemmap_page(struct page *page)
+{
+ if (PageReserved(page))
+ free_bootmem_page(page);
+ else
+ __free_page(page);
+}
+
+/* Free a list of the vmemmap pages */
+static void free_vmemmap_page_list(struct list_head *list)
+{
+ struct page *page, *next;
+
+ list_for_each_entry_safe(page, next, list, lru) {
+ list_del(&page->lru);
+ free_vmemmap_page(page);
+ }
+}
+
+static void vmemmap_remap_pte(pte_t *pte, unsigned long addr,
+ struct vmemmap_remap_walk *walk)
+{
+ /*
+ * Remap the tail pages as read-only to catch illegal write operation
+ * to the tail pages.
+ */
+ pgprot_t pgprot = PAGE_KERNEL_RO;
+ pte_t entry = mk_pte(walk->reuse_page, pgprot);
+ struct page *page = pte_page(*pte);
+
+ list_add(&page->lru, walk->vmemmap_pages);
+ set_pte_at(&init_mm, addr, pte, entry);
+}
+
+/**
+ * vmemmap_remap_free - remap the vmemmap virtual address range [@start, @end)
+ * to the page which @reuse is mapped to, then free vmemmap
+ * which the range are mapped to.
+ * @start: start address of the vmemmap virtual address range that we want
+ * to remap.
+ * @end: end address of the vmemmap virtual address range that we want to
+ * remap.
+ * @reuse: reuse address.
+ *
+ * Note: This function depends on vmemmap being base page mapped. Please make
+ * sure that the architecture disables PMD mapping of vmemmap pages when calling
+ * this function.
+ */
+void vmemmap_remap_free(unsigned long start, unsigned long end,
+ unsigned long reuse)
+{
+ LIST_HEAD(vmemmap_pages);
+ struct vmemmap_remap_walk walk = {
+ .remap_pte = vmemmap_remap_pte,
+ .reuse_addr = reuse,
+ .vmemmap_pages = &vmemmap_pages,
+ };
+
+ /*
+ * In order to make remapping routine most efficient for the huge pages,
+ * the routine of vmemmap page table walking has the following rules
+ * (see more details from the vmemmap_pte_range()):
+ *
+ * - The range [@start, @end) and the range [@reuse, @reuse + PAGE_SIZE)
+ * should be continuous.
+ * - The @reuse address is part of the range [@reuse, @end) that we are
+ * walking which is passed to vmemmap_remap_range().
+ * - The @reuse address is the first in the complete range.
+ *
+ * So we need to make sure that @start and @reuse meet the above rules.
+ */
+ BUG_ON(start - reuse != PAGE_SIZE);
+
+ vmemmap_remap_range(reuse, end, &walk);
+ free_vmemmap_page_list(&vmemmap_pages);
+}

/*
* Allocate a block of memory to be used to back the virtual memory map
diff --git a/include/linux/hugetlb_cgroup.h b/include/linux/hugetlb_cgroup.h
index 2ad6e92f124a..c0cae6a704f2 100644
--- a/include/linux/hugetlb_cgroup.h
+++ b/include/linux/hugetlb_cgroup.h
@@ -21,15 +21,16 @@ struct hugetlb_cgroup;
struct resv_map;
struct file_region;

+#ifdef CONFIG_CGROUP_HUGETLB
/*
* Minimum page order trackable by hugetlb cgroup.
* At least 4 pages are necessary for all the tracking information.
* The second tail page (hpage[2]) is the fault usage cgroup.
* The third tail page (hpage[3]) is the reservation usage cgroup.
*/
-#define HUGETLB_CGROUP_MIN_ORDER 2
+#define HUGETLB_CGROUP_MIN_ORDER order_base_2(NR_USED_SUBPAGE)

-#ifdef CONFIG_CGROUP_HUGETLB
enum hugetlb_memory_event {
HUGETLB_MAX,
HUGETLB_NR_MEMORY_EVENTS,
@@ -66,9 +67,9 @@ __hugetlb_cgroup_from_page(struct page *page, bool rsvd)
if (compound_order(page) < HUGETLB_CGROUP_MIN_ORDER)
return NULL;
if (rsvd)
- return (struct hugetlb_cgroup *)page[3].private;
+ return (void *)page_private(page + SUBPAGE_INDEX_CGROUP_RSVD);
else
- return (struct hugetlb_cgroup *)page[2].private;
+ return (void *)page_private(page + SUBPAGE_INDEX_CGROUP);
}

static inline struct hugetlb_cgroup *hugetlb_cgroup_from_page(struct page *page)
@@ -90,9 +91,11 @@ static inline int __set_hugetlb_cgroup(struct page *page,
if (compound_order(page) < HUGETLB_CGROUP_MIN_ORDER)
return -1;
if (rsvd)
- page[3].private = (unsigned long)h_cg;
+ set_page_private(page + SUBPAGE_INDEX_CGROUP_RSVD,
+ (unsigned long)h_cg);
else
- page[2].private = (unsigned long)h_cg;
+ set_page_private(page + SUBPAGE_INDEX_CGROUP,
+ (unsigned long)h_cg);
return 0;
}
@@ -21,15 +21,16 @@ struct hugetlb_cgroup;
* Minimum page order trackable by hugetlb cgroup.
* At least 4 pages are necessary for all the tracking information.
- * The second tail page (hpage[2]) is the fault usage cgroup.
- * The third tail page (hpage[3]) is the reservation usage cgroup.
+ * The second tail page (hpage[SUBPAGE_INDEX_CGROUP]) is the fault
+ * usage cgroup. The third tail page (hpage[SUBPAGE_INDEX_CGROUP_RSVD])
+ * is the reservation usage cgroup.
*/

--
2.11.0

2021-02-05 00:02:14

by Miaohe Lin

[permalink] [raw]
Subject: Re: [PATCH v14 2/8] mm: hugetlb: introduce a new config HUGETLB_PAGE_FREE_VMEMMAP

Hi:
On 2021/2/4 11:50, Muchun Song wrote:
> The option HUGETLB_PAGE_FREE_VMEMMAP allows for the freeing of
> some vmemmap pages associated with pre-allocated HugeTLB pages.
> For example, on X86_64 6 vmemmap pages of size 4KB each can be
> saved for each 2MB HugeTLB page. 4094 vmemmap pages of size 4KB
> each can be saved for each 1GB HugeTLB page.
>
> When a HugeTLB page is allocated or freed, the vmemmap array
> representing the range associated with the page will need to be
> remapped. When a page is allocated, vmemmap pages are freed
> after remapping. When a page is freed, previously discarded
> vmemmap pages must be allocated before remapping.
>
> The config option is introduced early so that supporting code
> can be written to depend on the option. The initial version of
> the code only provides support for x86-64.
>
> Like other code which frees vmemmap, this config option depends on
> HAVE_BOOTMEM_INFO_NODE. The routine register_page_bootmem_info() is
> used to register bootmem info. Therefore, make sure
> register_page_bootmem_info is enabled if HUGETLB_PAGE_FREE_VMEMMAP
> is defined.
>
> Signed-off-by: Muchun Song <[email protected]>
> Reviewed-by: Oscar Salvador <[email protected]>
> Acked-by: Mike Kravetz <[email protected]>

LGTM. Thanks.
Reviewed-by: Miaohe Lin <[email protected]>

> ---
> arch/x86/mm/init_64.c | 2 +-
> fs/Kconfig | 6 ++++++
> 2 files changed, 7 insertions(+), 1 deletion(-)
>
> diff --git a/arch/x86/mm/init_64.c b/arch/x86/mm/init_64.c
> index 0a45f062826e..0435bee2e172 100644
> --- a/arch/x86/mm/init_64.c
> +++ b/arch/x86/mm/init_64.c
> @@ -1225,7 +1225,7 @@ static struct kcore_list kcore_vsyscall;
>
> static void __init register_page_bootmem_info(void)
> {
> -#ifdef CONFIG_NUMA
> +#if defined(CONFIG_NUMA) || defined(CONFIG_HUGETLB_PAGE_FREE_VMEMMAP)
> int i;
>
> for_each_online_node(i)
> diff --git a/fs/Kconfig b/fs/Kconfig
> index 97e7b77c9309..de87f234f1e9 100644
> --- a/fs/Kconfig
> +++ b/fs/Kconfig
> @@ -237,6 +237,12 @@ config HUGETLBFS
> config HUGETLB_PAGE
> def_bool HUGETLBFS
>
> +config HUGETLB_PAGE_FREE_VMEMMAP
> + def_bool HUGETLB_PAGE
> + depends on X86_64
> + depends on SPARSEMEM_VMEMMAP
> + depends on HAVE_BOOTMEM_INFO_NODE
> +
> config MEMFD_CREATE
> def_bool TMPFS || HUGETLBFS
>
>

2021-02-05 07:35:36

by Miaohe Lin

[permalink] [raw]
Subject: Re: [PATCH v14 7/8] mm: hugetlb: gather discrete indexes of tail page

On 2021/2/4 11:50, Muchun Song wrote:
> For HugeTLB page, there are more metadata to save in the struct page.
> But the head struct page cannot meet our needs, so we have to abuse
> other tail struct page to store the metadata. In order to avoid
> conflicts caused by subsequent use of more tail struct pages, we can
> gather these discrete indexes of tail struct page. In this case, it
> will be easier to add a new tail page index later.
>
> There are only (RESERVE_VMEMMAP_SIZE / sizeof(struct page)) struct
> page structs that can be used when CONFIG_HUGETLB_PAGE_FREE_VMEMMAP,
> so add a BUILD_BUG_ON to catch invalid usage of the tail struct page.
>
> Signed-off-by: Muchun Song <[email protected]>
> Reviewed-by: Oscar Salvador <[email protected]>

Thanks.
Reviewed-by: Miaohe Lin <[email protected]>

> ---
> include/linux/hugetlb.h | 20 ++++++++++++++++++--
> include/linux/hugetlb_cgroup.h | 19 +++++++++++--------
> mm/hugetlb_vmemmap.c | 8 ++++++++
> 3 files changed, 37 insertions(+), 10 deletions(-)
>
> diff --git a/include/linux/hugetlb.h b/include/linux/hugetlb.h
> index 775aea53669a..822ab2f5542a 100644
> --- a/include/linux/hugetlb.h
> +++ b/include/linux/hugetlb.h
> @@ -28,6 +28,22 @@ typedef struct { unsigned long pd; } hugepd_t;
> #include <linux/shm.h>
> #include <asm/tlbflush.h>
>
> +/*
> + * For HugeTLB page, there are more metadata to save in the struct page. But
> + * the head struct page cannot meet our needs, so we have to abuse other tail
> + * struct page to store the metadata. In order to avoid conflicts caused by
> + * subsequent use of more tail struct pages, we gather these discrete indexes
> + * of tail struct page here.
> + */
> +enum {
> + SUBPAGE_INDEX_SUBPOOL = 1, /* reuse page->private */
> +#ifdef CONFIG_CGROUP_HUGETLB
> + SUBPAGE_INDEX_CGROUP, /* reuse page->private */
> + SUBPAGE_INDEX_CGROUP_RSVD, /* reuse page->private */
> +#endif
> + NR_USED_SUBPAGE,
> +};
> +
> struct hugepage_subpool {
> spinlock_t lock;
> long count;
> @@ -607,13 +623,13 @@ extern unsigned int default_hstate_idx;
> */
> static inline struct hugepage_subpool *hugetlb_page_subpool(struct page *hpage)
> {
> - return (struct hugepage_subpool *)(hpage+1)->private;
> + return (void *)page_private(hpage + SUBPAGE_INDEX_SUBPOOL);
> }
>
> static inline void hugetlb_set_page_subpool(struct page *hpage,
> struct hugepage_subpool *subpool)
> {
> - set_page_private(hpage+1, (unsigned long)subpool);
> + set_page_private(hpage + SUBPAGE_INDEX_SUBPOOL, (unsigned long)subpool);
> }
>
> static inline struct hstate *hstate_file(struct file *f)
> diff --git a/include/linux/hugetlb_cgroup.h b/include/linux/hugetlb_cgroup.h
> index 2ad6e92f124a..c0cae6a704f2 100644
> --- a/include/linux/hugetlb_cgroup.h
> +++ b/include/linux/hugetlb_cgroup.h
> @@ -21,15 +21,16 @@ struct hugetlb_cgroup;
> struct resv_map;
> struct file_region;
>
> +#ifdef CONFIG_CGROUP_HUGETLB
> /*
> * Minimum page order trackable by hugetlb cgroup.
> * At least 4 pages are necessary for all the tracking information.
> - * The second tail page (hpage[2]) is the fault usage cgroup.
> - * The third tail page (hpage[3]) is the reservation usage cgroup.
> + * The second tail page (hpage[SUBPAGE_INDEX_CGROUP]) is the fault
> + * usage cgroup. The third tail page (hpage[SUBPAGE_INDEX_CGROUP_RSVD])
> + * is the reservation usage cgroup.
> */
> -#define HUGETLB_CGROUP_MIN_ORDER 2
> +#define HUGETLB_CGROUP_MIN_ORDER order_base_2(NR_USED_SUBPAGE)
>
> -#ifdef CONFIG_CGROUP_HUGETLB
> enum hugetlb_memory_event {
> HUGETLB_MAX,
> HUGETLB_NR_MEMORY_EVENTS,
> @@ -66,9 +67,9 @@ __hugetlb_cgroup_from_page(struct page *page, bool rsvd)
> if (compound_order(page) < HUGETLB_CGROUP_MIN_ORDER)
> return NULL;
> if (rsvd)
> - return (struct hugetlb_cgroup *)page[3].private;
> + return (void *)page_private(page + SUBPAGE_INDEX_CGROUP_RSVD);
> else
> - return (struct hugetlb_cgroup *)page[2].private;
> + return (void *)page_private(page + SUBPAGE_INDEX_CGROUP);
> }
>
> static inline struct hugetlb_cgroup *hugetlb_cgroup_from_page(struct page *page)
> @@ -90,9 +91,11 @@ static inline int __set_hugetlb_cgroup(struct page *page,
> if (compound_order(page) < HUGETLB_CGROUP_MIN_ORDER)
> return -1;
> if (rsvd)
> - page[3].private = (unsigned long)h_cg;
> + set_page_private(page + SUBPAGE_INDEX_CGROUP_RSVD,
> + (unsigned long)h_cg);
> else
> - page[2].private = (unsigned long)h_cg;
> + set_page_private(page + SUBPAGE_INDEX_CGROUP,
> + (unsigned long)h_cg);
> return 0;
> }
>
> diff --git a/mm/hugetlb_vmemmap.c b/mm/hugetlb_vmemmap.c
> index 36ebd677e606..8efad9978821 100644
> --- a/mm/hugetlb_vmemmap.c
> +++ b/mm/hugetlb_vmemmap.c
> @@ -272,6 +272,14 @@ void __init hugetlb_vmemmap_init(struct hstate *h)
> unsigned int nr_pages = pages_per_huge_page(h);
> unsigned int vmemmap_pages;
>
> + /*
> + * There are only (RESERVE_VMEMMAP_SIZE / sizeof(struct page)) struct
> + * page structs that can be used when CONFIG_HUGETLB_PAGE_FREE_VMEMMAP,
> + * so add a BUILD_BUG_ON to catch invalid usage of the tail struct page.
> + */
> + BUILD_BUG_ON(NR_USED_SUBPAGE >=
> + RESERVE_VMEMMAP_SIZE / sizeof(struct page));
> +
> if (!hugetlb_free_vmemmap_enabled)
> return;
>
>

2021-02-05 08:57:02

by Oscar Salvador

[permalink] [raw]
Subject: Re: [PATCH v14 3/8] mm: hugetlb: free the vmemmap pages associated with each HugeTLB page

On Thu, Feb 04, 2021 at 11:50:38AM +0800, Muchun Song wrote:
> Every HugeTLB has more than one struct page structure. We __know__ that
> we only use the first 4(HUGETLB_CGROUP_MIN_ORDER) struct page structures
> to store metadata associated with each HugeTLB.
>
> There are a lot of struct page structures associated with each HugeTLB
> page. For tail pages, the value of compound_head is the same. So we can
> reuse first page of tail page structures. We map the virtual addresses
> of the remaining pages of tail page structures to the first tail page
> struct, and then free these page frames. Therefore, we need to reserve
> two pages as vmemmap areas.
>
> When we allocate a HugeTLB page from the buddy, we can free some vmemmap
> pages associated with each HugeTLB page. It is more appropriate to do it
> in the prep_new_huge_page().
>
> The free_vmemmap_pages_per_hpage(), which indicates how many vmemmap
> pages associated with a HugeTLB page can be freed, returns zero for
> now, which means the feature is disabled. We will enable it once all
> the infrastructure is there.
>
> Signed-off-by: Muchun Song <[email protected]>
> ---

[...]

> +void free_huge_page_vmemmap(struct hstate *h, struct page *head)
> +{
> + unsigned long vmemmap_addr = (unsigned long)head;
> + unsigned long vmemmap_end, vmemmap_reuse;
> +
> + if (!free_vmemmap_pages_per_hpage(h))
> + return;
> +
> + vmemmap_addr += RESERVE_VMEMMAP_SIZE;
> + vmemmap_end = vmemmap_addr + free_vmemmap_pages_size_per_hpage(h);
> + vmemmap_reuse = vmemmap_addr - PAGE_SIZE;
> +
> + /*
> + * Remap the vmemmap virtual address range [@vmemmap_addr, @vmemmap_end)
> + * to the page which @vmemmap_reuse is mapped to, then free the vmemmap
> + * pages which the range are mapped to.

"then free the pages which the range [@vmemmap_addr, @vmemmap_end] is mapped to."

I am not a native but sounds better to me.

> + */
> + vmemmap_remap_free(vmemmap_addr, vmemmap_end, vmemmap_reuse);
> +}
> diff --git a/mm/hugetlb_vmemmap.h b/mm/hugetlb_vmemmap.h
> new file mode 100644
> index 000000000000..6923f03534d5
> --- /dev/null
> +++ b/mm/hugetlb_vmemmap.h

[...]

> diff --git a/mm/sparse-vmemmap.c b/mm/sparse-vmemmap.c
> index 16183d85a7d5..50c1dc00b686 100644
> --- a/mm/sparse-vmemmap.c
> +++ b/mm/sparse-vmemmap.c
> @@ -27,8 +27,215 @@
> #include <linux/spinlock.h>
> #include <linux/vmalloc.h>
> #include <linux/sched.h>
> +#include <linux/pgtable.h>
> +#include <linux/bootmem_info.h>
> +
> #include <asm/dma.h>
> #include <asm/pgalloc.h>
> +#include <asm/tlbflush.h>
> +
> +/**
> + * vmemmap_remap_walk - walk vmemmap page table
> + *
> + * @remap_pte: called for each non-empty PTE (lowest-level) entry.

Well, we BUG_ON on empty PTE, so not sure that pointing out here is worth.
It sounds like we do nothing when it's empty.
Maybe:

"called for each lowest-level entry (PTE)"

> + * @reuse_page: the page which is reused for the tail vmemmap pages.
> + * @reuse_addr: the virtual address of the @reuse_page page.
> + * @vmemmap_pages: the list head of the vmemmap pages that can be freed.
> + */
> +struct vmemmap_remap_walk {
> + void (*remap_pte)(pte_t *pte, unsigned long addr,
> + struct vmemmap_remap_walk *walk);
> + struct page *reuse_page;
> + unsigned long reuse_addr;
> + struct list_head *vmemmap_pages;
> +};
> +
> +static void vmemmap_pte_range(pmd_t *pmd, unsigned long addr,
> + unsigned long end,
> + struct vmemmap_remap_walk *walk)
> +{
> + pte_t *pte;
> +
> + pte = pte_offset_kernel(pmd, addr);
> +
> + /*
> + * The reuse_page is found 'first' in table walk before we start
> + * remapping (which is calling @walk->remap_pte).
> + */
> + if (!walk->reuse_page) {
> + BUG_ON(pte_none(*pte) || walk->reuse_addr != addr);

I would rather have them in separate lines:
BUG_ON(pte_none(*pte));
BUG_ON(walk->reuse_addr != addr));

It helps when trying to figure out when we explode. One could dig in the
registers, but let's make it easier to find out.

> +

[...]


> +static void vmemmap_remap_range(unsigned long start, unsigned long end,
> + struct vmemmap_remap_walk *walk)
> +{
> + unsigned long addr = start;
> + unsigned long next;
> + pgd_t *pgd;
> +
> + VM_BUG_ON(!IS_ALIGNED(start, PAGE_SIZE));
> + VM_BUG_ON(!IS_ALIGNED(end, PAGE_SIZE));
> +
> + pgd = pgd_offset_k(addr);
> + do {
> + BUG_ON(pgd_none(*pgd));
> +
> + next = pgd_addr_end(addr, end);
> + vmemmap_p4d_range(pgd, addr, next, walk);
> + } while (pgd++, addr = next, addr != end);
> +
> + /*
> + * We do not change the mapping of the vmemmap virtual address range
> + * [@start, @start + PAGE_SIZE) which belongs to the reuse range.
> + * So we not need to flush the TLB.
> + */
> + flush_tlb_kernel_range(start + PAGE_SIZE, end);

I find that comment a bit confusing. I would rather describe what are we
flushing instead of what we are not.


> +}
> +
> +/*
> + * Free a vmemmap page. A vmemmap page can be allocated from the memblock
> + * allocator or buddy allocator. If the PG_reserved flag is set, it means
> + * that it allocated from the memblock allocator, just free it via the
> + * free_bootmem_page(). Otherwise, use __free_page().
> + */
> +static inline void free_vmemmap_page(struct page *page)
> +{
> + if (PageReserved(page))
> + free_bootmem_page(page);
> + else
> + __free_page(page);
> +}
> +
> +/* Free a list of the vmemmap pages */
> +static void free_vmemmap_page_list(struct list_head *list)
> +{
> + struct page *page, *next;
> +
> + list_for_each_entry_safe(page, next, list, lru) {
> + list_del(&page->lru);
> + free_vmemmap_page(page);
> + }
> +}
> +
> +static void vmemmap_remap_pte(pte_t *pte, unsigned long addr,
> + struct vmemmap_remap_walk *walk)
> +{
> + /*
> + * Remap the tail pages as read-only to catch illegal write operation
> + * to the tail pages.
> + */
> + pgprot_t pgprot = PAGE_KERNEL_RO;
> + pte_t entry = mk_pte(walk->reuse_page, pgprot);
> + struct page *page = pte_page(*pte);
> +
> + list_add(&page->lru, walk->vmemmap_pages);
> + set_pte_at(&init_mm, addr, pte, entry);
> +}
> +
> +/**
> + * vmemmap_remap_free - remap the vmemmap virtual address range [@start, @end)
> + * to the page which @reuse is mapped to, then free vmemmap
> + * which the range are mapped to.
> + * @start: start address of the vmemmap virtual address range that we want
> + * to remap.
> + * @end: end address of the vmemmap virtual address range that we want to
> + * remap.
> + * @reuse: reuse address.
> + *
> + * Note: This function depends on vmemmap being base page mapped. Please make
> + * sure that the architecture disables PMD mapping of vmemmap pages when calling
> + * this function.

Well, we do not really depend on the architecture to not map the vmemmap range
with PMDs, right? IIUC, that is driven by your boot parameter (patch#5), which
overrides whatever the architecture can do.

Functional changes look good to me, so with all the above fixes, you can add:

Reviewed-by: Oscar Salvador <[email protected]>


--
Oscar Salvador
SUSE L3

2021-02-05 09:07:43

by Oscar Salvador

[permalink] [raw]
Subject: Re: [PATCH v14 0/8] Free some vmemmap pages of HugeTLB page

On Thu, Feb 04, 2021 at 11:50:35AM +0800, Muchun Song wrote:
> Changelog in v13 -> v14:
> - Refuse to free the HugeTLB page when the system is under memory pressure.
> - Use GFP_ATOMIC to allocate vmemmap pages instead of GFP_KERNEL.
> - Rebase to linux-next 20210202.
> - Fix and add some comments for vmemmap_remap_free().

What has happened to [1] ? AFAIK we still need it, right?
If not, could you explain why?

[1] https://patchwork.kernel.org/project/linux-mm/patch/[email protected]/

>
> Thanks to Oscar, Mike, David H and David R's suggestions and review.
>
> Changelog in v12 -> v13:
> - Remove VM_WARN_ON_PAGE macro.
> - Add more comments in vmemmap_pte_range() and vmemmap_remap_free().
>
> Thanks to Oscar and Mike's suggestions and review.
>
> Changelog in v11 -> v12:
> - Move VM_WARN_ON_PAGE to a separate patch.
> - Call __free_hugepage() with hugetlb_lock (See patch #5.) to serialize
> with dissolve_free_huge_page(). It is to prepare for patch #9.
> - Introduce PageHugeInflight. See patch #9.
>
> Changelog in v10 -> v11:
> - Fix compiler error when !CONFIG_HUGETLB_PAGE_FREE_VMEMMAP.
> - Rework some comments and commit changes.
> - Rework vmemmap_remap_free() to 3 parameters.
>
> Thanks to Oscar and Mike's suggestions and review.
>
> Changelog in v9 -> v10:
> - Fix a bug in patch #11. Thanks to Oscar for pointing that out.
> - Rework some commit log or comments. Thanks Mike and Oscar for the suggestions.
> - Drop VMEMMAP_TAIL_PAGE_REUSE in the patch #3.
>
> Thank you very much Mike and Oscar for reviewing the code.
>
> Changelog in v8 -> v9:
> - Rework some code. Very thanks to Oscar.
> - Put all the non-hugetlb vmemmap functions under sparsemem-vmemmap.c.
>
> Changelog in v7 -> v8:
> - Adjust the order of patches.
>
> Very thanks to David and Oscar. Your suggestions are very valuable.
>
> Changelog in v6 -> v7:
> - Rebase to linux-next 20201130
> - Do not use basepage mapping for vmemmap when this feature is disabled.
> - Rework some patchs.
> [PATCH v6 08/16] mm/hugetlb: Free the vmemmap pages associated with each hugetlb page
> [PATCH v6 10/16] mm/hugetlb: Allocate the vmemmap pages associated with each hugetlb page
>
> Thanks to Oscar and Barry.
>
> Changelog in v5 -> v6:
> - Disable PMD/huge page mapping of vmemmap if this feature was enabled.
> - Simplify the first version code.
>
> Changelog in v4 -> v5:
> - Rework somme comments and code in the [PATCH v4 04/21] and [PATCH v4 05/21].
>
> Thanks to Mike and Oscar's suggestions.
>
> Changelog in v3 -> v4:
> - Move all the vmemmap functions to hugetlb_vmemmap.c.
> - Make the CONFIG_HUGETLB_PAGE_FREE_VMEMMAP default to y, if we want to
> disable this feature, we should disable it by a boot/kernel command line.
> - Remove vmemmap_pgtable_{init, deposit, withdraw}() helper functions.
> - Initialize page table lock for vmemmap through core_initcall mechanism.
>
> Thanks for Mike and Oscar's suggestions.
>
> Changelog in v2 -> v3:
> - Rename some helps function name. Thanks Mike.
> - Rework some code. Thanks Mike and Oscar.
> - Remap the tail vmemmap page with PAGE_KERNEL_RO instead of PAGE_KERNEL.
> Thanks Matthew.
> - Add some overhead analysis in the cover letter.
> - Use vmemap pmd table lock instead of a hugetlb specific global lock.
>
> Changelog in v1 -> v2:
> - Fix do not call dissolve_compound_page in alloc_huge_page_vmemmap().
> - Fix some typo and code style problems.
> - Remove unused handle_vmemmap_fault().
> - Merge some commits to one commit suggested by Mike.
>
> Muchun Song (8):
> mm: memory_hotplug: factor out bootmem core functions to
> bootmem_info.c
> mm: hugetlb: introduce a new config HUGETLB_PAGE_FREE_VMEMMAP
> mm: hugetlb: free the vmemmap pages associated with each HugeTLB page
> mm: hugetlb: alloc the vmemmap pages associated with each HugeTLB page
> mm: hugetlb: add a kernel parameter hugetlb_free_vmemmap
> mm: hugetlb: introduce nr_free_vmemmap_pages in the struct hstate
> mm: hugetlb: gather discrete indexes of tail page
> mm: hugetlb: optimize the code with the help of the compiler
>
> Documentation/admin-guide/kernel-parameters.txt | 14 ++
> Documentation/admin-guide/mm/hugetlbpage.rst | 3 +
> arch/x86/mm/init_64.c | 13 +-
> fs/Kconfig | 6 +
> include/linux/bootmem_info.h | 65 +++++
> include/linux/hugetlb.h | 43 +++-
> include/linux/hugetlb_cgroup.h | 19 +-
> include/linux/memory_hotplug.h | 27 --
> include/linux/mm.h | 5 +
> mm/Makefile | 2 +
> mm/bootmem_info.c | 124 ++++++++++
> mm/hugetlb.c | 23 +-
> mm/hugetlb_vmemmap.c | 314 ++++++++++++++++++++++++
> mm/hugetlb_vmemmap.h | 33 +++
> mm/memory_hotplug.c | 116 ---------
> mm/sparse-vmemmap.c | 280 +++++++++++++++++++++
> mm/sparse.c | 1 +
> 17 files changed, 930 insertions(+), 158 deletions(-)
> create mode 100644 include/linux/bootmem_info.h
> create mode 100644 mm/bootmem_info.c
> create mode 100644 mm/hugetlb_vmemmap.c
> create mode 100644 mm/hugetlb_vmemmap.h
>
> --
> 2.11.0
>
>

--
Oscar Salvador
SUSE L3

2021-02-05 09:39:07

by Muchun Song

[permalink] [raw]
Subject: Re: [External] Re: [PATCH v14 0/8] Free some vmemmap pages of HugeTLB page

On Fri, Feb 5, 2021 at 4:59 PM Oscar Salvador <[email protected]> wrote:
>
> On Thu, Feb 04, 2021 at 11:50:35AM +0800, Muchun Song wrote:
> > Changelog in v13 -> v14:
> > - Refuse to free the HugeTLB page when the system is under memory pressure.
> > - Use GFP_ATOMIC to allocate vmemmap pages instead of GFP_KERNEL.
> > - Rebase to linux-next 20210202.
> > - Fix and add some comments for vmemmap_remap_free().
>
> What has happened to [1] ? AFAIK we still need it, right?
> If not, could you explain why?
>
> [1] https://patchwork.kernel.org/project/linux-mm/patch/[email protected]/

Hi Oscar,

I reply to you in another thread (in the patch #4).

Thanks. :-)

>
> >
> > Thanks to Oscar, Mike, David H and David R's suggestions and review.
> >
> > Changelog in v12 -> v13:
> > - Remove VM_WARN_ON_PAGE macro.
> > - Add more comments in vmemmap_pte_range() and vmemmap_remap_free().
> >
> > Thanks to Oscar and Mike's suggestions and review.
> >
> > Changelog in v11 -> v12:
> > - Move VM_WARN_ON_PAGE to a separate patch.
> > - Call __free_hugepage() with hugetlb_lock (See patch #5.) to serialize
> > with dissolve_free_huge_page(). It is to prepare for patch #9.
> > - Introduce PageHugeInflight. See patch #9.
> >
> > Changelog in v10 -> v11:
> > - Fix compiler error when !CONFIG_HUGETLB_PAGE_FREE_VMEMMAP.
> > - Rework some comments and commit changes.
> > - Rework vmemmap_remap_free() to 3 parameters.
> >
> > Thanks to Oscar and Mike's suggestions and review.
> >
> > Changelog in v9 -> v10:
> > - Fix a bug in patch #11. Thanks to Oscar for pointing that out.
> > - Rework some commit log or comments. Thanks Mike and Oscar for the suggestions.
> > - Drop VMEMMAP_TAIL_PAGE_REUSE in the patch #3.
> >
> > Thank you very much Mike and Oscar for reviewing the code.
> >
> > Changelog in v8 -> v9:
> > - Rework some code. Very thanks to Oscar.
> > - Put all the non-hugetlb vmemmap functions under sparsemem-vmemmap.c.
> >
> > Changelog in v7 -> v8:
> > - Adjust the order of patches.
> >
> > Very thanks to David and Oscar. Your suggestions are very valuable.
> >
> > Changelog in v6 -> v7:
> > - Rebase to linux-next 20201130
> > - Do not use basepage mapping for vmemmap when this feature is disabled.
> > - Rework some patchs.
> > [PATCH v6 08/16] mm/hugetlb: Free the vmemmap pages associated with each hugetlb page
> > [PATCH v6 10/16] mm/hugetlb: Allocate the vmemmap pages associated with each hugetlb page
> >
> > Thanks to Oscar and Barry.
> >
> > Changelog in v5 -> v6:
> > - Disable PMD/huge page mapping of vmemmap if this feature was enabled.
> > - Simplify the first version code.
> >
> > Changelog in v4 -> v5:
> > - Rework somme comments and code in the [PATCH v4 04/21] and [PATCH v4 05/21].
> >
> > Thanks to Mike and Oscar's suggestions.
> >
> > Changelog in v3 -> v4:
> > - Move all the vmemmap functions to hugetlb_vmemmap.c.
> > - Make the CONFIG_HUGETLB_PAGE_FREE_VMEMMAP default to y, if we want to
> > disable this feature, we should disable it by a boot/kernel command line.
> > - Remove vmemmap_pgtable_{init, deposit, withdraw}() helper functions.
> > - Initialize page table lock for vmemmap through core_initcall mechanism.
> >
> > Thanks for Mike and Oscar's suggestions.
> >
> > Changelog in v2 -> v3:
> > - Rename some helps function name. Thanks Mike.
> > - Rework some code. Thanks Mike and Oscar.
> > - Remap the tail vmemmap page with PAGE_KERNEL_RO instead of PAGE_KERNEL.
> > Thanks Matthew.
> > - Add some overhead analysis in the cover letter.
> > - Use vmemap pmd table lock instead of a hugetlb specific global lock.
> >
> > Changelog in v1 -> v2:
> > - Fix do not call dissolve_compound_page in alloc_huge_page_vmemmap().
> > - Fix some typo and code style problems.
> > - Remove unused handle_vmemmap_fault().
> > - Merge some commits to one commit suggested by Mike.
> >
> > Muchun Song (8):
> > mm: memory_hotplug: factor out bootmem core functions to
> > bootmem_info.c
> > mm: hugetlb: introduce a new config HUGETLB_PAGE_FREE_VMEMMAP
> > mm: hugetlb: free the vmemmap pages associated with each HugeTLB page
> > mm: hugetlb: alloc the vmemmap pages associated with each HugeTLB page
> > mm: hugetlb: add a kernel parameter hugetlb_free_vmemmap
> > mm: hugetlb: introduce nr_free_vmemmap_pages in the struct hstate
> > mm: hugetlb: gather discrete indexes of tail page
> > mm: hugetlb: optimize the code with the help of the compiler
> >
> > Documentation/admin-guide/kernel-parameters.txt | 14 ++
> > Documentation/admin-guide/mm/hugetlbpage.rst | 3 +
> > arch/x86/mm/init_64.c | 13 +-
> > fs/Kconfig | 6 +
> > include/linux/bootmem_info.h | 65 +++++
> > include/linux/hugetlb.h | 43 +++-
> > include/linux/hugetlb_cgroup.h | 19 +-
> > include/linux/memory_hotplug.h | 27 --
> > include/linux/mm.h | 5 +
> > mm/Makefile | 2 +
> > mm/bootmem_info.c | 124 ++++++++++
> > mm/hugetlb.c | 23 +-
> > mm/hugetlb_vmemmap.c | 314 ++++++++++++++++++++++++
> > mm/hugetlb_vmemmap.h | 33 +++
> > mm/memory_hotplug.c | 116 ---------
> > mm/sparse-vmemmap.c | 280 +++++++++++++++++++++
> > mm/sparse.c | 1 +
> > 17 files changed, 930 insertions(+), 158 deletions(-)
> > create mode 100644 include/linux/bootmem_info.h
> > create mode 100644 mm/bootmem_info.c
> > create mode 100644 mm/hugetlb_vmemmap.c
> > create mode 100644 mm/hugetlb_vmemmap.h
> >
> > --
> > 2.11.0
> >
> >
>
> --
> Oscar Salvador
> SUSE L3

2021-02-05 23:24:37

by Muchun Song

[permalink] [raw]
Subject: Re: [External] Re: [PATCH v14 3/8] mm: hugetlb: free the vmemmap pages associated with each HugeTLB page

On Fri, Feb 5, 2021 at 4:54 PM Oscar Salvador <[email protected]> wrote:
>
> On Thu, Feb 04, 2021 at 11:50:38AM +0800, Muchun Song wrote:
> > Every HugeTLB has more than one struct page structure. We __know__ that
> > we only use the first 4(HUGETLB_CGROUP_MIN_ORDER) struct page structures
> > to store metadata associated with each HugeTLB.
> >
> > There are a lot of struct page structures associated with each HugeTLB
> > page. For tail pages, the value of compound_head is the same. So we can
> > reuse first page of tail page structures. We map the virtual addresses
> > of the remaining pages of tail page structures to the first tail page
> > struct, and then free these page frames. Therefore, we need to reserve
> > two pages as vmemmap areas.
> >
> > When we allocate a HugeTLB page from the buddy, we can free some vmemmap
> > pages associated with each HugeTLB page. It is more appropriate to do it
> > in the prep_new_huge_page().
> >
> > The free_vmemmap_pages_per_hpage(), which indicates how many vmemmap
> > pages associated with a HugeTLB page can be freed, returns zero for
> > now, which means the feature is disabled. We will enable it once all
> > the infrastructure is there.
> >
> > Signed-off-by: Muchun Song <[email protected]>
> > ---
>
> [...]
>
> > +void free_huge_page_vmemmap(struct hstate *h, struct page *head)
> > +{
> > + unsigned long vmemmap_addr = (unsigned long)head;
> > + unsigned long vmemmap_end, vmemmap_reuse;
> > +
> > + if (!free_vmemmap_pages_per_hpage(h))
> > + return;
> > +
> > + vmemmap_addr += RESERVE_VMEMMAP_SIZE;
> > + vmemmap_end = vmemmap_addr + free_vmemmap_pages_size_per_hpage(h);
> > + vmemmap_reuse = vmemmap_addr - PAGE_SIZE;
> > +
> > + /*
> > + * Remap the vmemmap virtual address range [@vmemmap_addr, @vmemmap_end)
> > + * to the page which @vmemmap_reuse is mapped to, then free the vmemmap
> > + * pages which the range are mapped to.
>
> "then free the pages which the range [@vmemmap_addr, @vmemmap_end] is mapped to."
>
> I am not a native but sounds better to me.

Me too. But I believe you are right. :-)

>
> > + */
> > + vmemmap_remap_free(vmemmap_addr, vmemmap_end, vmemmap_reuse);
> > +}
> > diff --git a/mm/hugetlb_vmemmap.h b/mm/hugetlb_vmemmap.h
> > new file mode 100644
> > index 000000000000..6923f03534d5
> > --- /dev/null
> > +++ b/mm/hugetlb_vmemmap.h
>
> [...]
>
> > diff --git a/mm/sparse-vmemmap.c b/mm/sparse-vmemmap.c
> > index 16183d85a7d5..50c1dc00b686 100644
> > --- a/mm/sparse-vmemmap.c
> > +++ b/mm/sparse-vmemmap.c
> > @@ -27,8 +27,215 @@
> > #include <linux/spinlock.h>
> > #include <linux/vmalloc.h>
> > #include <linux/sched.h>
> > +#include <linux/pgtable.h>
> > +#include <linux/bootmem_info.h>
> > +
> > #include <asm/dma.h>
> > #include <asm/pgalloc.h>
> > +#include <asm/tlbflush.h>
> > +
> > +/**
> > + * vmemmap_remap_walk - walk vmemmap page table
> > + *
> > + * @remap_pte: called for each non-empty PTE (lowest-level) entry.
>
> Well, we BUG_ON on empty PTE, so not sure that pointing out here is worth.
> It sounds like we do nothing when it's empty.
> Maybe:
>
> "called for each lowest-level entry (PTE)"

Thanks. I will update this.

>
> > + * @reuse_page: the page which is reused for the tail vmemmap pages.
> > + * @reuse_addr: the virtual address of the @reuse_page page.
> > + * @vmemmap_pages: the list head of the vmemmap pages that can be freed.
> > + */
> > +struct vmemmap_remap_walk {
> > + void (*remap_pte)(pte_t *pte, unsigned long addr,
> > + struct vmemmap_remap_walk *walk);
> > + struct page *reuse_page;
> > + unsigned long reuse_addr;
> > + struct list_head *vmemmap_pages;
> > +};
> > +
> > +static void vmemmap_pte_range(pmd_t *pmd, unsigned long addr,
> > + unsigned long end,
> > + struct vmemmap_remap_walk *walk)
> > +{
> > + pte_t *pte;
> > +
> > + pte = pte_offset_kernel(pmd, addr);
> > +
> > + /*
> > + * The reuse_page is found 'first' in table walk before we start
> > + * remapping (which is calling @walk->remap_pte).
> > + */
> > + if (!walk->reuse_page) {
> > + BUG_ON(pte_none(*pte) || walk->reuse_addr != addr);
>
> I would rather have them in separate lines:
> BUG_ON(pte_none(*pte));
> BUG_ON(walk->reuse_addr != addr));
>
> It helps when trying to figure out when we explode. One could dig in the
> registers, but let's make it easier to find out.

OK. Will do.

>
> > +
>
> [...]
>
>
> > +static void vmemmap_remap_range(unsigned long start, unsigned long end,
> > + struct vmemmap_remap_walk *walk)
> > +{
> > + unsigned long addr = start;
> > + unsigned long next;
> > + pgd_t *pgd;
> > +
> > + VM_BUG_ON(!IS_ALIGNED(start, PAGE_SIZE));
> > + VM_BUG_ON(!IS_ALIGNED(end, PAGE_SIZE));
> > +
> > + pgd = pgd_offset_k(addr);
> > + do {
> > + BUG_ON(pgd_none(*pgd));
> > +
> > + next = pgd_addr_end(addr, end);
> > + vmemmap_p4d_range(pgd, addr, next, walk);
> > + } while (pgd++, addr = next, addr != end);
> > +
> > + /*
> > + * We do not change the mapping of the vmemmap virtual address range
> > + * [@start, @start + PAGE_SIZE) which belongs to the reuse range.
> > + * So we not need to flush the TLB.
> > + */
> > + flush_tlb_kernel_range(start + PAGE_SIZE, end);
>
> I find that comment a bit confusing. I would rather describe what are we
> flushing instead of what we are not.
>

OK. Will update it.

>
> > +}
> > +
> > +/*
> > + * Free a vmemmap page. A vmemmap page can be allocated from the memblock
> > + * allocator or buddy allocator. If the PG_reserved flag is set, it means
> > + * that it allocated from the memblock allocator, just free it via the
> > + * free_bootmem_page(). Otherwise, use __free_page().
> > + */
> > +static inline void free_vmemmap_page(struct page *page)
> > +{
> > + if (PageReserved(page))
> > + free_bootmem_page(page);
> > + else
> > + __free_page(page);
> > +}
> > +
> > +/* Free a list of the vmemmap pages */
> > +static void free_vmemmap_page_list(struct list_head *list)
> > +{
> > + struct page *page, *next;
> > +
> > + list_for_each_entry_safe(page, next, list, lru) {
> > + list_del(&page->lru);
> > + free_vmemmap_page(page);
> > + }
> > +}
> > +
> > +static void vmemmap_remap_pte(pte_t *pte, unsigned long addr,
> > + struct vmemmap_remap_walk *walk)
> > +{
> > + /*
> > + * Remap the tail pages as read-only to catch illegal write operation
> > + * to the tail pages.
> > + */
> > + pgprot_t pgprot = PAGE_KERNEL_RO;
> > + pte_t entry = mk_pte(walk->reuse_page, pgprot);
> > + struct page *page = pte_page(*pte);
> > +
> > + list_add(&page->lru, walk->vmemmap_pages);
> > + set_pte_at(&init_mm, addr, pte, entry);
> > +}
> > +
> > +/**
> > + * vmemmap_remap_free - remap the vmemmap virtual address range [@start, @end)
> > + * to the page which @reuse is mapped to, then free vmemmap
> > + * which the range are mapped to.
> > + * @start: start address of the vmemmap virtual address range that we want
> > + * to remap.
> > + * @end: end address of the vmemmap virtual address range that we want to
> > + * remap.
> > + * @reuse: reuse address.
> > + *
> > + * Note: This function depends on vmemmap being base page mapped. Please make
> > + * sure that the architecture disables PMD mapping of vmemmap pages when calling
> > + * this function.
>
> Well, we do not really depend on the architecture to not map the vmemmap range
> with PMDs, right? IIUC, that is driven by your boot parameter (patch#5), which
> overrides whatever the architecture can do.

Right. I will rework the comment here.

>
> Functional changes look good to me, so with all the above fixes, you can add:
>
> Reviewed-by: Oscar Salvador <[email protected]>
>

Very thanks.

>
> --
> Oscar Salvador
> SUSE L3

2021-02-05 23:29:57

by Joao Martins

[permalink] [raw]
Subject: Re: [PATCH v14 0/8] Free some vmemmap pages of HugeTLB page

On 2/4/21 3:50 AM, Muchun Song wrote:
> Hi all,
>

[...]

> When a HugeTLB is freed to the buddy system, we should allocate 6 pages for
> vmemmap pages and restore the previous mapping relationship.
>
> Apart from 2MB HugeTLB page, we also have 1GB HugeTLB page. It is similar
> to the 2MB HugeTLB page. We also can use this approach to free the vmemmap
> pages.
>
> In this case, for the 1GB HugeTLB page, we can save 4094 pages. This is a
> very substantial gain. On our server, run some SPDK/QEMU applications which
> will use 1024GB hugetlbpage. With this feature enabled, we can save ~16GB
> (1G hugepage)/~12GB (2MB hugepage) memory.
>
> Because there are vmemmap page tables reconstruction on the freeing/allocating
> path, it increases some overhead. Here are some overhead analysis.

[...]

> Although the overhead has increased, the overhead is not significant. Like Mike
> said, "However, remember that the majority of use cases create hugetlb pages at
> or shortly after boot time and add them to the pool. So, additional overhead is
> at pool creation time. There is no change to 'normal run time' operations of
> getting a page from or returning a page to the pool (think page fault/unmap)".
>

Despite the overhead and in addition to the memory gains from this series ...
there's an additional benefit there isn't talked here with your vmemmap page
reuse trick. That is page (un)pinners will see an improvement and I presume because
there are fewer memmap pages and thus the tail/head pages are staying in cache more
often.

Out of the box I saw (when comparing linux-next against linux-next + this series)
with gup_test and pinning a 16G hugetlb file (with 1G pages):

get_user_pages(): ~32k -> ~9k
unpin_user_pages(): ~75k -> ~70k

Usually any tight loop fetching compound_head(), or reading tail pages data (e.g.
compound_head) benefit a lot. There's some unpinning inefficiencies I am fixing[0], but
with that in added it shows even more:

unpin_user_pages(): ~27k -> ~3.8k

FWIW, I was also seeing that with devdax and the ZONE_DEVICE vmemmap page reuse equivalent
series[1] but it was mixed with other numbers.

Anyways, JFYI :)

Joao

[0] https://lore.kernel.org/linux-mm/[email protected]/
[1] https://lore.kernel.org/linux-mm/[email protected]/